#!/usr/bin/python

import string
import os
import getopt
import sys
import re

prefix='@prefix@'
datadir='@datadir@'
libdir="@libdir@"

# run from textrace-source dir.

autotrace_bin = ''
if libdir == "@" + "libdir" + "@":
	libdir = os.path.join (os.getcwd (), 'autotrace/')

autotrace_bin = os.path.join (libdir, "autotrace")

one_em = 1000
version='0.48'
map_file = ''
trace_file = ''
valid_pbm_file = ''


if datadir == '@' + "datadir" + "@":
	datadir = os.getcwd ()

trace_file = os.path.join (datadir, 'trace2.ps')
gsx_translator = os.path.join (datadir, 't1d2gsx.pl')
type1fix_file = os.path.join (datadir, 'type1fix.pl')
valid_xbm_file = os.path.join (datadir, 'valid.xbm')

valid_xbm = open (valid_xbm_file).read ()

def system (cmd):
	sys.stderr.write ("Invoking `%s'\n"% cmd)
	return os.system (cmd)

def success_system (cmd):
	st = system(cmd)
	if st:
		sys.stderr.write("Command exited with status %d\n" % st)
		sys.exit (2)
		
def run_autotrace  (fi, fn, opts):
	opts = " " + opts + " --output-format=bps --input-format=pbm "

	stat = system (autotrace_bin + opts + " --output-file=char.eps -filter-iterations %d %s.pbm  " % (fi,fn) )
	return stat

def do_autotrace_optimal_fi (fn, opts):
	"""
Run autotrace, and find the best  filter-iterations value using binary search.

	"""
	hi = 9
	lo = 0
	
	if run_autotrace (hi, fn, opts) ==0:
		return 1

	if run_autotrace (lo, fn, opts) <> 0 :
		return 0

	while 1:
		mid = (lo + hi) /2
		if mid == lo or mid == hi:
			break
		
		if run_autotrace (mid, fn, opts) == 0:
			lo = mid
		else:
			hi = mid
		
	run_autotrace (lo, fn, opts)

def blank_pbm (filename):
	"""
Kill the contents of a PBM: write 0xFF to the PBM.
"""
	f = open (filename)
	l = f.readline ()
	length = len (f.read())

	f.close ()
	
	open (filename, 'w').write ('%s\n%s' % (l,'\377' * length))

def do_autotrace_one (pbmfile):
	"""
	Run autotrace, first with -background-color, then without.
"""
	bg = do_autotrace_optimal_fi (pbmfile , '-background-color FFFFFF')
	if not bg:
		bg = do_autotrace_optimal_fi (pbmfile,'')

	if not bg:
		sys.stderr.write ("Failing even without -background-color\n")
		# todo: copy .PBM file to `date`+`time`.pbm.

		blank_pbm (pbmfile + ".xbm")
		if run_autotrace (9, '-background-color FFFFFF') <> 0:
			raise 'Flabbergasted'
	

def compose_tfm_number (seq):
	shift = (len (seq)-1)*8

	cs = 0
	for b in seq:
		cs = cs  + (ord (b) << shift)
		shift = shift - 8
	return cs

def extract_size (tfmfont):
	f = open (tfmfont).read ()

	design_size = compose_tfm_number (f[12* 2 + 4: 12*2 + 8])
	design_size = (design_size /16.0) / (1 << 16)

	return design_size
	

def make_xbms (fontname):
	" Extract bitmaps from the PK file using `pk2bm'. "
	
	sys.stderr.write ('Making bitmaps...')
	sys.stderr.flush() 

	glyphs = []
	for a in range(0,256):
		oldxbm = os.popen ("pk2bm -o %o %s" % (a, fontname)).read ()
		if len (oldxbm) == 0:
			oldxbm = valid_xbm

		sys.stderr.write ('[%d]' % a)
		src = '_%c_' % chr (a)
		oldxbm = string.replace (oldxbm, src, '%d_' % a)
		os.popen ("xbmtopbm > %d.pbm" % (a), 'w').write (oldxbm)

	sys.stderr.write ('done.\n')
	sys.stderr.flush() 

def get_glyph_info (fontname):
	fn = {}
	ch = 0
	for l in os.popen ('pktype %s' % fontname).readlines ():
		m = re.search ('Character = ([0-9]+)', l)
		if m :
			ch = string.atoi (m.group (1))
			continue
		m = re.search ('Height = ([0-9]+).*Width = ([0-9]+).*X-offset = ([-0-9]+).*Y-offset = ([0-9-]+)', l)
		if m:
			dims = map (string.atoi, m.groups ())
			fn[ch] = tuple (dims)

	return fn

def write_dimen_file (font_dict, glyphs):
	f = open ("dimension-file", 'w')

	for a  in glyphs:
		if font_dict.has_key (a):
			(h, w, xo, yo) =  font_dict[a]

			wd = w
			ht = yo
			dp = h - yo
			stx = xo
			sty = dp
			
			f.write ('whdx %d %d %d %d %d %d endwhdx\n' % (a, wd, ht, dp, stx, sty))
		else:
			f.write ('whdx %d 20 10 10 0 10 endwhdx\n' % a) 
	
	f.close()

def setup_gs_handler ():
	open ('error-handler.ps', 'w').write('''/handleerror {
    (error-output) (w) file
    dup (Fatal: GS error: ) writestring
    dup $error /errorname get write==only
    dup ( in ) writestring
    dup $error /command get write==only
    dup ( at ) writestring
    dup $error /position get write==only
    dup (.\n) writestring
    flushfile flush quit
    } bind def''')
	
def autotrace_font (fontname_base, glyphs):
	fontname =  gen_pk (fontname_base)
	make_xbms (fontname)

	fn = get_glyph_info (fontname)
	write_dimen_file (fn, glyphs);

	setup_gs_handler ()

	try:
		os.unlink ('first.t1d')
	except:
		pass
	
	for a in glyphs:
		do_autotrace_one ("%d" % a) 
		system ('gs -d_n="%d" -dNODISPLAY -q error-handler.ps %s >> first.t1d' % (a, trace_file))

	to_type1 (fontname_base)

def to_type1 (fontname):
	preamble = '''
FontName =	TeX-%s
VVV =		001
WWW =		001
# Notice =	by (pts@fazekas.hu) at Wed Apr  4 19:41:13 CEST 2001
# Copyright =	Copyright by (pts@fazekas.hu) at Wed Apr  4 19:41:13 CEST 2001
Notice =	converted after April 2001
Copyright =	see copyright of original TeX font
FullName =	TeX %s Regular
FamilyName =	TeX %s
Weight =	Regular
ItalicAngle =	0
isFixedPitch =	false
UnderlinePosition = -100
UnderlineThickness = 50
# UniqueID =	4313101
UniqueID = 	%d
xrevscale =	%f
yrevscale =	%f

Encoding=
CharStrings=
CharBBox=
''' % (fontname, fontname, fontname, 0, 0.001, 0.001)
	body = open ("first.t1d"). read ()
	body = preamble + body
	open ("second.t1d",'w').write (body)
	success_system ("perl %s --256 < second.t1d > first.gsx" % gsx_translator)
	success_system ('perl -x %s first.gsx pfb: %s --pack --dump-spaces=no --debug-warnings --chk-insize=no' %(type1fix_file, fontname))

#	success_system ("perl -x  %s first.gsx pfa: pfafile.pfa  --pack --dump-spaces=no --debug-warnings --chk-insize=no" % type1fix_file)

	
def gen_pk (fontname):
	base_dpi = 600
	mag = 1

	tfm = os.popen ('kpsewhich %s.tfm' % fontname).read ()[:-1]
	size = extract_size (tfm)

	size_points = size * 1/72.27 * base_dpi
	
	mag = 1000.0 / size_points
	prod = mag * base_dpi
	try:
		f = open ('%s.%dpk' % (fontname, prod))
	except:
		system(r"mf '\mode:=ljfour; mag:=%f; nonstopmode; input %s'" % (mag,fontname))
		log = open ('%s.log' % fontname).read ()
		m = re.search('Output written on %s.([0-9]+)gf' % fontname, log)
		prod = string.atoi (m.group (1))
		
		system("gftopk %s.%dgf" % (fontname, prod))

	
	return "%s.%dpk" % (fontname , prod)

startdir = os.getcwd()
try:
	os.mkdir('temp')
except:
	pass

#try:
#	os.link (sys.argv[1], 'temp/')
#except:
#	pass

glyph_range = range(0,256)

(opts,files) = getopt.getopt (sys.argv[1:], 'r:',['range='])


for (o,a) in opts:
	if o == '--range' or o == '-r':
		bs = map (string.atoi , string.split (a, ','))
		glyph_range = range (bs[0], bs[1])
		


os.chdir ('temp')
autotrace_font (files[0], glyph_range)
